/*
|  A simple multi-threaded HTTP server.
|
| THANKS:
| - socket library overview-> http://www.linuxhowtos.org/C_C++/socket.htm
| - strnstr() implementation-> http://stackoverflow.com/questions/23999797/implementing-strnstr
|    ^ Creative Commons license.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <errno.h>
#include <sys/stat.h>
#include <pthread.h>
#include <signal.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <limits.h>
#include <sys/random.h>
#include "hmac.c"
#include "sha1.c"
#include "strnstr.c"      // necessary for many Linux systems.
                          // (BSD varients often implement strnstr() in the standard C lib.)
#include <sys/select.h>

/* _____ OPTIONS _____ */
#define NUM_THREADS 8       // number of threads, number of simultaneous requests handled.
                            /* (Because no connection is persistant, a higher number will result
                               in increased performance for a single client and for multiple clients,
                               up to a point. Optimal number would be the number of hardware threads. ) */

#define BUFFER_SIZE 1024    // size of client buffer in bytes, minimum 512

#define LOGIN_PAGE "login.html"    // route to the admin login interface
#define ADMIN_PAGE "admin.txt"     // route to the admin dashboard
#define PASSWORD_CHARS "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
#define PASSWORD_LEN 20
const uint8_t* cookie_key = (char []){
    0x4E, 0x2B, 0xE2, 0x5A, 0xCC, 0x99, 0xA9, 0xAA,
    0xC6, 0x6F, 0xB6, 0xCE, 0x21, 0x7F, 0x07, 0x1A,
    0x47, 0x0B, 0xCE, 0x95, 0xB5, 0x3D, 0x6E, 0xD5,
    0xB4, 0x6D, 0xA7, 0xE9, 0xF4, 0xD6, 0x55, 0x5E };    // 32 byte key


static int STATUS_NONE = -1;

/* _____ GLOBALS _____ */
static int port;                                               // port number
static long addr;
static char* dir;                                              // root directory
static int sock, cli_addr_len, n[NUM_THREADS];                 // socket handle;
static int connections[NUM_THREADS];                           // connection handles
static char* buffers[NUM_THREADS];                             // connection buffers
static struct sockaddr_in serv_addr,cli_addr[NUM_THREADS];     // server and client addresses
static int online;                                             // status of server
static pthread_t threads[NUM_THREADS];                         // thread handles
static int avail[NUM_THREADS];                                 // client id's
static pthread_mutex_t mutex;                                  // thread syncronization
static pthread_mutex_t logfile;                                // safely write to the log
char password[PASSWORD_LEN+1];                                 // admin password


/* _____ DATA _____ */
typedef struct {
    uint32_t authenticated;    // must be exactly 1 to be authenticated
    time_t expiration;         // must be in the future to be authenticated
} cookie_data;


/* _____ FUNCTIONS _____ */
void error(char *msg);                          // error handler
int add_client();                               // allocate client id
void sub_client();                              // deallocate client id
void parse_args(int argc, const char* argv[] ); // parse command line arguments
uint32_t uniform_random(uint32_t upper_bound);  // generate a nonnegative integer less than upper_bound uniformly at random
void generate_password();                       // create and print the admin password
void prepare_socket();                          // create & bind socket
void accept_connections();                      // listen for & deploy connections to threads
void* serve(void* client);                      // serve client (thread workload)
char* get_header(int id,
                 char* packet,
                 int hostname_len);             // generate header for response
char* header_helper(int id,
                    int status_code,
                    char* status_msg,
                    char* content_type,
                    int packet_length,
                    int filename_length);       // prints header to thread's buffer
char* cookie_header_helper(int id, char* url);
char* redirect_helper(int id, char* url);       // prints status 301 header to thread's buffer
void log_request(char* packet);                 // write a short summary of the request to admin.txt
void send_header(int connection, char* header); // send header to client
void send_data(int connection,
               int file,
               char* dynamic_page);             // send data from (templated) file to client
int file_open(char* file);                      // open file for reading
int file_size(char* file);                      // get file size in bytes
int file_exists(char* file);                    // check if file exists
void file_build_path(int id, char* packet);     // combine root directory w/ requested file's path in thread's buffer
char* file_build_absolute_path(int id, char* filename); // write filename to beginning of buffer
char* url_build(int id, char* packet);           // build url in thread's buffer
void go_offline();                              // first step in exit--break all loops
int is_online();                                // check online status
int packet_is_admin_page(char* packet);         // determine whether request is for the admin page
int packet_is_login_page(char* packet);         // determine whether request is for the login page
int file_is_admin_page(char* file);             // determine whether request is for the admin page
int file_is_login_page(char* file);             // determine whether request is for the login page
int auth(char* packet);                         // check if username and password are valid admin credentials
int make_cookie(uint8_t* buffer);               // generates a new cookie, encodes it in base16, and writes it to buffer
int check_cookie(char* packet);                 // check if cookie is valid for the admin page
void cleanup_and_exit(int);                        // go_offline, cleanup memory, join threads, and exit (called w/ CTRL+C)

int main(int argc, const char * argv[])
{
    /* parse args */
    parse_args(argc, argv);

    /* generate admin password */
    generate_password();
    /* register exit signal (CTRL+C) */
    signal(SIGINT, cleanup_and_exit);

    /* initialize mutex */
    pthread_mutex_init(&mutex, NULL);
    pthread_mutex_init(&logfile, NULL);

    /* initialize client ID's */
    for( int i = 0; i < NUM_THREADS; i++){
        avail[i] = STATUS_NONE;
    }
    /* prepare socket */
    prepare_socket();

    /* accept connections (loop) */
    accept_connections();
}

/* parse CLI arguments */
void parse_args(int argc, const char* argv[] )
{
    // check number of args
    if( argc != 4 ) error("usage: webserver <ip> <port #> <root directory>\n");

    addr = inet_addr(argv[1]);
    // check if port in range
    port = atoi(argv[2]);
    if( port == 0 ) error("port number must be greater than zero.\n");

    // save root directory
    dir = malloc(strlen(argv[3])+1);
    strcpy(dir, argv[3]);

    // verify dir exists
    if (0 != access(dir, F_OK)) {
        if (ENOENT == errno) {
            // does not exist
            error("Root directory does not exist");
        }
        if (ENOTDIR == errno) {
            // not a directory
            error("Path given is not directory");
        }
    }
}

/* generate a nonnegative integer less than upper_bound uniformly at random */
uint32_t uniform_random(uint32_t upper_bound)
{
    if( upper_bound == 0 ) exit(EXIT_FAILURE);
    if( upper_bound == 1 ) return 1;
    uint32_t mask = 1;
    while( mask < upper_bound ){
        mask = 2*(mask+1)-1; // mask is always of the form 2^n-1
    }
    uint32_t bytes[4];
    while(1){
        getrandom(bytes, 4*4, 0);    // 4 bytes in a uint32_t
        for( int i=0; i < 4; i++ ){
            uint32_t candidate = bytes[i] & mask;
            if( candidate < upper_bound ){
                return candidate;
            }
        }
    };
}

/* create and print the admin password */
void generate_password()
{
    for( int i=0; i<PASSWORD_LEN; i++ ){
        uint32_t random = uniform_random(strlen(PASSWORD_CHARS));
        password[i] = PASSWORD_CHARS[random];
    }
    password[PASSWORD_LEN] = '\0';
    printf("\n------------------------------\nAdmin password follows, keep it secret:\n%s\n", password);
}

/* create & bind socket */
void prepare_socket()
{
    /* create socket */
    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &(int){1}, sizeof(int)) < 0)
      error("setsockopt(SO_REUSEADDR) failed");
    if( sock < 0 ) error("Couldn't create socket");

    /* prepare tools */
    bzero((char *) &serv_addr, sizeof(serv_addr));  // initialize to zero
    serv_addr.sin_family = AF_INET;                 // set address family to 'internet'
    serv_addr.sin_port = htons(port);               // convert host port to network port
    serv_addr.sin_addr.s_addr = addr;

    /* bind socket */
    if( bind(sock, (struct sockaddr*)&serv_addr, sizeof(serv_addr)) < 0 ){
        error("Coulnd't bind socket");}

    /* go online */
    listen(sock, 5);    // 5 is the standard maximum for waiting socket clients
    online = 1;
    printf("------------------------------\n    Listening on port %d\n       EXIT WITH CTRL+C       \n------------------------------\r\n",port);
}

/* accept socket connections, spin up thread and hand off session to thread */
void accept_connections()
{
    cli_addr_len = sizeof(cli_addr);
    while(online){
      int id = add_client();
      if( id != STATUS_NONE ){
            bzero(&cli_addr[id], sizeof(struct sockaddr_in));
            connections[id] = accept(sock,
                                     (struct sockaddr*) &cli_addr[id],
                                     (unsigned int *)&cli_addr_len);
            if ( connections[id] < 0 ){ error("Error accepting client.\n"); }
            if ( pthread_create(&threads[id], NULL, serve, &avail[id]) != 0 ){
                error("Couldn't create thread");
            }

        }
    }
}

long int get_packet_from_client(int id) {
    buffers[id] = calloc(sizeof(char), BUFFER_SIZE);
    int total_buffer_size = BUFFER_SIZE;
    int packet_size = 0;


    struct timeval timeout;
    timeout.tv_sec = 0;
    timeout.tv_usec = 0;

    fd_set readfds;
    FD_ZERO(&readfds);
    FD_SET(connections[id], &readfds);
    char temp_buf[BUFFER_SIZE];

    while(1){

        int temp_size = (int)read(connections[id], temp_buf, BUFFER_SIZE);
        if ( temp_size < 0 ){
            return temp_size;
        }
        if ( packet_size + temp_size > total_buffer_size ){
            total_buffer_size *= 2;
            buffers[id] = (char *) realloc(buffers[id], total_buffer_size);
        }
        strncpy(buffers[id] + packet_size, temp_buf, temp_size);
        packet_size += temp_size;

        int sel_rv = select(connections[id] + 1, &readfds, NULL, NULL, &timeout);
        if ( sel_rv == 0 ){
            break;
        }

        if ( temp_size == 0 ){
            break;
        }
    }
    return packet_size;
}

static short int MAX_BUFFER_SIZE = 4000;

/* thread workload -- respond to client requests */
void* serve(void* arg)
{
    /* allocate resources */
    int id = *((int*)arg);                  // local copy of id makes following code cleaner
    int timeout = 0;                        // timeout counter; 3 1-second empty reads until session ends

    /* serving client */
    while( is_online() && timeout < 3 )
    {
        n[id] = get_packet_from_client(id);

        // if there's nothing to read
        if( n[id] <= 0 ){ timeout++; /*increment timeout counter*/ }
        else{
            /* save host name from buffer */
            char* end = strnstr(buffers[id], "Host: ", 256);    // find beginning of host name
            if (end == NULL){ // host not supplied
                break; // protect against file name too large or doesnt contain host
            }
            char* start = end + 6;                              // most past "Host: "
            end = strnstr(start, "\r", 32);                     // find end of host name
            if ( end == NULL ){
                break;
            }
            int host_len = (int)(end - start);                  // find length of host name
            char host[host_len+1];                              // create storage
            strncpy(host, start, host_len);                     // copy name
            host[host_len] = '\0';                              // add nullchar (strncpy doesn't)

            log_request(buffers[id]);
            char* header = get_header(id, buffers[id], host_len);
            send_header(id, header);     // parse what was read & send response header
            if ( MAX_BUFFER_SIZE < (short int) n[id] ){                 // check input is good size
                send_data(id, file_open("400.html"), NULL);
                break;
            }
            char temp_buf[MAX_BUFFER_SIZE];                     // put in local buffer
            strncpy(temp_buf, buffers[id], n[id]);

            if( strnstr(temp_buf, "http://",7) == NULL ){            // if not a redirect
                if( file_is_login_page(temp_buf) ){
                    send_data(id, file_open(LOGIN_PAGE), NULL);
                }
                else if( file_is_admin_page(temp_buf) ){             // already authenticated via get_header
                    send_data(id, file_open(ADMIN_PAGE), NULL);
                }
                else if( file_exists(temp_buf)){
                    send_data(id, file_open(temp_buf), NULL);          // send HTTP payload
                }
                else{   // if file doesn't exist
                    char url[100]; // url used in templated 404 page
                    snprintf(url, 100, "http://%s%s",host,temp_buf+strlen(dir));
                    send_data(id, file_open("404.html"), url);
                }
            }
            /* NO CONNECTIONS ARE PERSISTANT--if response was sent, close connection */
            break;
        }
    }

    /* ending session */
    shutdown(connections[id], 2);   // close connection
    free(buffers[id]);              // deallocate buffer
    sub_client(id);                 // deallocate client ID
    pthread_detach(threads[id]);    // detach thread to deallocate resources
    return NULL;
}

/* build response header based on request packet */
char* get_header(int id, char* packet, int host_len)
{
    char* request = strnstr(packet, "GET", 32);     // search for GET
    if( request != NULL )
    {
        /* if /go/ redirect specified */
        if( strnstr(packet, " /go/", 16) != NULL){
            return redirect_helper(id, url_build(id, packet));
        }

        /* if user is authenticating via login page */
        if( packet_is_login_page(packet) ){
            if( check_cookie(packet) || auth(packet) ){
                char url[32] = "/";
                strcat(url, ADMIN_PAGE);
                return cookie_header_helper(id, url);
            }
            else{
                file_build_absolute_path(id, LOGIN_PAGE);
                return header_helper(id, 200, "OK", "text/html", file_size(LOGIN_PAGE), (int)strlen(LOGIN_PAGE));
            }
        }

        /* if request is for admin page */
        else if( packet_is_admin_page(packet) ){
            if( check_cookie(packet) ){
                file_build_absolute_path(id, ADMIN_PAGE);
                return header_helper(id, 200, "OK", "text/plain", file_size(ADMIN_PAGE), (int)strlen(ADMIN_PAGE));
            }
            else{
                return redirect_helper(id, file_build_absolute_path(id, LOGIN_PAGE));
            }
        }

        /* if no page specified */
        else if( strnstr(packet, " / ", 16) != NULL ){      // default "/" to "/index.html"
          file_build_path(id, "/index.html ");
        }

        /* if page is specified */
        else if( strnstr(packet, " /", 16) != NULL ){
          file_build_path(id, packet);
        }

        /* verify file exists, then generate header */
        if( file_exists(buffers[id]) ){
            if( strnstr(buffers[id], ".jpg", 50) != NULL){
                return header_helper(id, 200, "OK", "image/jpeg", file_size(buffers[id]), (int)strlen(buffers[id]));
            }
            if( strnstr(buffers[id], ".png", 50) != NULL ){
                return header_helper(id, 200, "OK", "image/png", file_size(buffers[id]), (int)strlen(buffers[id]));
            }
            if( strnstr(buffers[id], ".html", 50) != NULL ){
                return header_helper(id, 200, "OK", "text/html", file_size(buffers[id]), (int)strlen(buffers[id]));
            }
            if( strnstr(buffers[id], ".css", 50) != NULL ){
                return header_helper(id, 200, "OK", "text/css", file_size(buffers[id]), (int)strlen(buffers[id]));
            }
            if( strnstr(buffers[id], ".js", 50) != NULL ){
                return header_helper(id, 200, "OK", "Application/javascript", file_size(buffers[id]), (int)strlen(buffers[id]));
            }
            return header_helper(id, 200,"OK", "text", file_size(buffers[id]), (int)strlen(buffers[id]));   // default text content
        }

        /* if file doesn't exist, generate 404 from template */
        else{
            int filename_len = (int)(strlen(buffers[id])-strlen(dir));  // length of filename without full dir path
            int url_len = 7;    // length of "http://"
            // <HTML> + "http://" + <host:port> + <file>
            int content_len = file_size("404.html") - 1
                              + url_len
                              + host_len
                              + filename_len;
            return header_helper(id, 404, "Not Found", "text/html", content_len, (int)strlen(buffers[id]));
        }
    }
    return NULL;    // shouldn't happen, but necessary for thread function prototype
                    // and useful for error detection.
}

/* print generic response header to buffer */
char* header_helper(int id, int status_code, char* status_msg, char* content_type, int content_length, int filename_length)
{
    char* header = buffers[id] + filename_length + 1;     // write header after file path in the buffer
    snprintf(header,
             120,   // limit on header length to avoid overflow in small buffers
             "HTTP/1.1 %d %s\r\n"
             "Content-Type: %s\r\n"
             "Content-Length: %d\r\n"
             "Connection: close\r\n"
             "\r\n", status_code, status_msg, content_type, content_length);
    return header;
}

char* cookie_header_helper(int id, char* url)
{
    char* header = buffers[id];
    snprintf(header,
             120,   // limit on header length to avoid overflow in small buffers
             "HTTP/1.1 301 Moved Permanently\r\n"
             "Location: %s\r\n"
             "Set-Cookie: session=",
             url);
    char* start = strnstr(header, "session=", 120)+8; // 8 characters in "session="
    int cookie_len = make_cookie(start);
    snprintf(start+cookie_len,
             120,
             "\r\n\r\n", url);
    return header;
}
/* print status 301 header in buffer for redirect functionality */
char* redirect_helper(int id, char* url)
{
    char* header = buffers[id] + strlen(url) + 1;     // write header after file path in the buffer
    snprintf(header,
             120,   // limit on header length to avoid overflow in small buffers
             "HTTP/1.1 301 Moved Permanently\r\n"
             "Location: %s\r\n"
             "\r\n", url);
    return header;
}

/* write a short summary of the request to admin.txt */
void log_request(char* packet)
{
    char to_log[163];
    char* end = strnstr(packet, "\r\n", 160);
    int len = (int)(end - packet);
    if( end == NULL ){
        len = (strlen(packet) > 159 ? 159 : strlen(packet));
    }
    // now len <= 159
    strncpy(to_log, packet, len);
    to_log[len] = '\n';
    to_log[len+1] = '\0';
    pthread_mutex_lock(&logfile);
    FILE* pFile = fopen(ADMIN_PAGE, "a");
    fputs(to_log, pFile);
    fclose(pFile);
    pthread_mutex_unlock(&logfile);
}

/* send header to client */
void send_header(int id, char* header)
{
    if( header != NULL ){
        int n = (int)strlen(header);
        void* pos = header;
        while (n > 0) {
            int bytes_written = (int)send(connections[id], pos, n, 0);
            if (bytes_written <= 0) { error("Couldn't send header."); }
            n -= bytes_written;
            pos += bytes_written;
        }
    }
}

/* Send input_file to client, optionally filling dynamic template data */
void send_data(int id, int input_file, char* dynamic)
{
    /* handle missing file cases */
    if( input_file == -1 ){
        input_file = open("404.html",O_RDONLY, S_IREAD);                        // 404 if specified file missing
        if( input_file == -1 ){ error("Necessary files missing: 404.html"); }   // error if 404 template missing
    }
    /* clear buffer */
    bzero(buffers[id], BUFFER_SIZE);

    /* send file to client */
    char* read_pos = buffers[id];
    char* dyn_pos = dynamic;
    while (1) {
        int bytes_read;
        if(dynamic == NULL){
            bytes_read = (int)read(input_file, buffers[id], BUFFER_SIZE);       // Read data into buffer.
        }
        else{
            bytes_read = (int)read(input_file, read_pos, 1);                    // Read data into buffer.
            read_pos++;
        }
        if (bytes_read == 0){ break; }                                      // Exit if nothing to read.
        if (bytes_read < 0){ error("Couldn't read file into buffer"); }     // Handle errors.
        if( dynamic != NULL ){
            /* render template */
            while ( bytes_read <= BUFFER_SIZE ) {                                       // while buffer has space
                int new_bytes = (int)read(input_file, read_pos, 1);                     // read 1 byte at a time
                if (new_bytes == 0){ break; }                                           // Exit if nothing to read.
                if (new_bytes < 0){ error("Couldn't read file into buffer"); }          // Handle errors.
                bytes_read += new_bytes;                                                // increment bytes_read
                if( *read_pos == '*' ){                                                 // if current char is template marker '*'
                    while( (*dyn_pos != '\0') && (bytes_read <= BUFFER_SIZE) ){         // insert dynamic data at marker
                        *read_pos = *dyn_pos;
                        read_pos++;
                        dyn_pos++;
                        bytes_read++;
                    }
                }
                else{
                    read_pos++;                                                         // if current != '*', continue reading
                }
            }
        }

        // Write data into socket.
        void *p = buffers[id];
        while (bytes_read > 0) {
            int bytes_written = (int)write(connections[id], p, bytes_read);
            if (bytes_written <= 0){ error("Couldn't send file to client."); }
            bytes_read -= bytes_written;
            p += bytes_written;
        }
    }

    /* close file */
    close(input_file);
}

/* open file for reading only */
int file_open(char* file)
{
    return open(file, O_RDONLY);
}

/* get file size in bytes */
int file_size(char* file)
{
    struct stat st;
    stat(file, &st);
    return (int)st.st_size;
}

/* check if file exists */
int file_exists(char* file)
{
    if (0 == access(file, 0)) {
        return 1;
    }
    return 0;
}

/* write filename to beginning of buffer */
char* file_build_absolute_path(int id, char* filename)
{
    int len = strlen(filename);                 // find length of filename
    bzero(buffers[id], BUFFER_SIZE);            // clear buffer
    strncpy(buffers[id], filename, len+1);      // add filename to path with null byte
    return buffers[id];
}

/* build root dir + filename path at beginning of buffer */
void file_build_path(int id, char* packet)
{
    char* start = strnstr(packet, "/", 100);    // find beginning of filename
    char* end = strnstr(start, " ", strlen(packet));       // find end of filename
    int len = (int)(end - start);               // find length of filename
    char filename[len+1];                       // allocate storage
    strncpy(filename, start, len);              // copy filename, buffer about to be wiped
    filename[len] = '\0';                       // add null (strncpy doesn't)
    bzero(buffers[id], strlen(buffers[id]) + strlen(dir));            // clear buffer
    strncpy(buffers[id], dir, strlen(dir));     // copy root dir to buffer
    strncpy(buffers[id]+strlen(dir), filename, len+1);    // add filename to path
}

/* builds URL of requested file based on packet's request line and "Host:" value */
char* url_build(int id, char* packet)
{
    char* start = strnstr(packet, "/go/", 100); // find beginning of hostname
    start += 4;                                  // move past "/go/"
    char* end = strnstr(start, " ", 100);       // find end of hostname
    int len = (int)(end - start);               // find length of filename
    char hostname[len+1];                       // allocate storage
    strncpy(hostname, start, len);              // copy hostname, buffer about to be wiped
    hostname[len] = '\0';                       // add null (strncpy doesn't)
    bzero(buffers[id], BUFFER_SIZE);            // clear buffer
    int url_components = 7 + 4 + 4 + 1;         // "http://" + "www." + ".com\0"
    snprintf(buffers[id], len+url_components, "http://www.%s.com",hostname);
    return buffers[id];
}

/* allocates client id */
int add_client()
{
    int e = 0;
    pthread_mutex_lock(&mutex);
    while( (avail[e] != STATUS_NONE) && (e < NUM_THREADS) ){    // find available id for client
        e++;
    }
    if( e < NUM_THREADS ){ avail[e] = e; }    // if id allocated, mark as used
    pthread_mutex_unlock(&mutex);
    if( e == NUM_THREADS ){
        return STATUS_NONE;
    }
    return e;
}

/* deallocates client id */
void sub_client(int id)
{
    pthread_mutex_lock(&mutex);
    avail[id] = STATUS_NONE;        // mark client ID as available
    pthread_mutex_unlock(&mutex);
}

/* deallocates client id */
int is_online()
{
    int x;
    pthread_mutex_lock(&mutex);
    x = online;
    pthread_mutex_unlock(&mutex);
    return x;
}

/* deallocates client id */
void go_offline()
{
    pthread_mutex_lock(&mutex);
    online = 0;
    pthread_mutex_unlock(&mutex);
}


/* returns the smallest non-NULL value from loc1 and loc2 */
char* earliest_occurrence(char* loc1, char* loc2)
{
    char* earliest;
    if( loc1 == NULL ){
        earliest = loc2;
    }
    else if( loc2 == NULL){
        earliest = loc1;
    }
    else {
        earliest = loc1 < loc2 ? loc1 : loc2;
    }
    return earliest;
}

/* determine whether request is for the admin page */
int packet_is_admin_page(char* packet)
{
    char* start = strnstr(packet, "/", 100)+1;    // find beginning of filename
    char* end = strnstr(start, " ", 100);         // find end of filename
    int len = (int)(end - start);                 // find length of filename
    if( len != strlen(ADMIN_PAGE) ) return 0;
    return !strncmp(start, ADMIN_PAGE, len);
}

/* determine whether request is for the login page */
/* Requests to the login page may have URL query parameters, so ignore those */
int packet_is_login_page(char* packet)
{
    char* start = strnstr(packet, "/", 100)+1;    // find beginning of filename
    char* end1 = strnstr(start, " ", 100);
    char* end2 = strnstr(start, "?", 100);
    char* end = earliest_occurrence(end1, end2);  // find end of filename, ignoring URL parameters
    int len = (int)(end - start);                 // find length of filename
    if( len != strlen(LOGIN_PAGE) ) return 0;
    return !strncmp(start, LOGIN_PAGE, len);
}

/* determine whether request is for the admin page */
int file_is_admin_page(char* file)
{
    return !strncmp(file, ADMIN_PAGE, 50);
}

/* determine whether request is for the login page */
int file_is_login_page(char* file)
{
    return !strncmp(file, LOGIN_PAGE, 50);
}

/* check if username and password are valid admin credentials */
int auth(char* packet)
{
    char* start = strnstr(packet, "?", 120);    // find beginning of querystring
    if( start == NULL ) return 0;               // no login details included
    char* end = strnstr(start, " ", 120)+1;     // find end of querystring
    int len = (int)(end - start);

    char* name_start = strnstr(start, "user=", len) + strlen("user=");
    if ( name_start == NULL ) return 0;         // no username
    char* name_end1 = strnstr(name_start, "&", len);
    char* name_end2 = strnstr(name_start, " ", len);
    char* name_end = earliest_occurrence(name_end1, name_end2);
    int name_len = (int)(name_end - name_start);

    char* password_start = strnstr(start, "password=", len) + strlen("password=");
    if ( password_start == NULL ) return 0;     // no password
    char* password_end1 = strnstr(password_start, "&", len);
    char* password_end2 = strnstr(password_start, " ", len);
    char* password_end = earliest_occurrence(password_end1, password_end2);
    int input_password_len = (int)(password_end - password_start);

    int name_cmp_len = name_len > strlen("admin") ? name_len : strlen("admin");
    int pass_cmp_len = input_password_len > PASSWORD_LEN ? input_password_len : PASSWORD_LEN;

    if( strncmp(name_start, "admin", name_cmp_len) != 0 ){           // Only user is "admin"
        return 0;
    }
    else if( strncmp(password_start, password, pass_cmp_len) != 0 ){ // Password must match
        return 0;
    }
    return 1;   // Authenticated
}

/* generates a new cookie, encodes it in base16, and writes it to buffer */
/* returns number of bytes written */
int make_cookie(uint8_t* buffer)
{
    time_t rawtime;
    time( &rawtime );
    cookie_data* to_sign = (cookie_data*)malloc(sizeof(cookie_data));
    to_sign->authenticated = 1;
    to_sign->expiration = rawtime + 60*60*48; // 48 hours in the future
    uint8_t signature[20];
    hmac_sha1(cookie_key, sizeof(cookie_key), (uint8_t *)to_sign, sizeof(cookie_data), (uint8_t *)signature);
    // write cookie data
    int data_size = sizeof(cookie_data);
    for( int i=0; i < data_size; i++ ){
        uint8_t byte = ((uint8_t*)to_sign)[i];
        snprintf(buffer+(2*i),3,"%02X",byte); // Write byte as ASCII, e.g. "9A" for 0x9A
    };
    // write signature
    for( int i=0; i < 20; i++ ){
        uint8_t byte = signature[i];
        snprintf(buffer+(2*data_size)+(2*i),3,"%02X",byte);
    };
    free(to_sign);
    return 2*data_size + 2*20;
}

/* check if cookie is valid for the admin page */
int check_cookie(char* packet)
{
    // Parse the header
    char* start = strnstr(packet, "session=", BUFFER_SIZE);
    if( start == NULL ) return 0; // no cookie
    start += 8; // 8 characters in "session="
    char* end = strnstr(start, "\r\n", 200);
    int len = (int)(end-start);
    if( len != 2*sizeof(cookie_data) + 2*20 ) return 0;

    // Get the cookie data
    uint8_t* cookie = (uint8_t*)malloc(sizeof(cookie_data));
    char* pos = start;
    for( int i=0; i<sizeof(cookie_data); i++ ){   // Get cookie_data
        sscanf(pos, "%2hhx", cookie+i);
        pos += 2;
    }

    // Get the signature from the cookie
    uint8_t given_signature[20];
    for( int i=0; i<20; i++ ){   // Get cookie_data
        sscanf(pos, "%2hhx", given_signature+i);
        pos += 2;
    }

    // Compare the cookie data
    time_t timenow;
    time( &timenow );
    if( ((cookie_data*)cookie)->expiration < timenow ||
        ((cookie_data*)cookie)->authenticated != 1 ){
        free(cookie);
        return 0;
    }

    // Compute the HMAC of cookie_data
    uint8_t computed_signature[20];
    hmac_sha1(cookie_key, sizeof(cookie_key), (uint8_t *)cookie, sizeof(cookie_data), (uint8_t *)computed_signature);
    free(cookie);

    // Compare the signatures
    for( int i=0; i<20; i++ ){
        if( computed_signature[i] != given_signature[i] ) return 0;
    }
    return 1;
}

/* Stop & join threads, close sockets, free malloc's, & exit. */
void cleanup_and_exit(int sig)
{
    printf("\nCleaning up...\n");
    go_offline();

    pthread_mutex_lock(&mutex);
    for( int i = 0; i < NUM_THREADS; i++ ){
      if (avail[i] != STATUS_NONE){
        pthread_join(threads[i], NULL);
        shutdown(connections[i], 2);
      }
    }
    pthread_mutex_unlock(&mutex);
    printf("\nthreads terminated...");
    shutdown(sock, 2);
    printf("\nsockets/connections closed...");
    free(dir);
    printf("\ngoodbye!\n");
    exit(0);
}

/* error handler */
void error(char *msg)
{
    perror(msg);
    exit(1);
}
